#define STRICT
#define WIN32_LEAN_AND_MEAN
#define _USE_MATH_DEFINES

#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <math.h>
#include "resource.h"

//-----------------------------------------------------------------------------
// GLOBALS
//-----------------------------------------------------------------------------
HWND                    g_hWnd          = NULL;
LPDIRECT3D9             g_pD3D          = NULL;
LPDIRECT3DDEVICE9       g_pd3dDevice    = NULL;
LPDIRECT3DVERTEXBUFFER9 g_pVertexBuffer = NULL;
LPDIRECT3DTEXTURE9      g_pFontTexture  = NULL;
LPDIRECT3DTEXTURE9      g_pDecalTexture = NULL;
LPDIRECT3DTEXTURE9      g_pBumpTexture = NULL;
LPD3DXEFFECT            g_pEffect       = NULL;

D3DXMATRIX g_matWorld;
D3DXMATRIX g_matView;
D3DXMATRIX g_matProj;


float g_LightDir[] = {1.0f, 0.0f, 0.0f };				// Light direction, passed to pixel shader
float g_TestCoordinate[] = {0.5f, 0.5f};				// Location of pixel to debug

float g_XRot = 0.0f;									// model X-rotation
float g_YRot = 0.0f;									// model Y-rotation

struct Vertex
{
    float x, y, z;
    float nx, ny, nz;
    float tu, tv;
};

Vertex g_quadVertices[] =
{
	{-1.0f, 1.0f, 0.0f,  0.0f, 0.0f, -1.0f,  0.0f,0.0f },
	{ 1.0f, 1.0f, 0.0f,  0.0f, 0.0f, -1.0f,  1.0f,0.0f },
	{-1.0f,-1.0f, 0.0f,  0.0f, 0.0f, -1.0f,  0.0f,1.0f },
	{ 1.0f,-1.0f, 0.0f,  0.0f, 0.0f, -1.0f,  1.0f,1.0f }
};

#define D3DFVF_VERTEX (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1)

//-----------------------------------------------------------------------------
// PROTOTYPES
//-----------------------------------------------------------------------------
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int nCmdShow);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
void init(void);
void render(void);
void shutDown(void);
void initEffect(void);
void setTechniqueVariables(void);

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR     lpCmdLine,
					int       nCmdShow )
{
	WNDCLASSEX winClass; 
	MSG        uMsg;

    memset(&uMsg,0,sizeof(uMsg));

	winClass.lpszClassName = "WINDOWS_CLASS";
	winClass.cbSize        = sizeof(WNDCLASSEX);
	winClass.style         = CS_HREDRAW | CS_VREDRAW;
	winClass.lpfnWndProc   = WindowProc;
	winClass.hInstance     = hInstance;
	winClass.hIcon	       = LoadIcon(hInstance, (LPCTSTR)IDI_DIRECTX_ICON);
    winClass.hIconSm	   = LoadIcon(hInstance, (LPCTSTR)IDI_DIRECTX_ICON);
	winClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	winClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winClass.lpszMenuName  = NULL;
	winClass.cbClsExtra    = 0;
	winClass.cbWndExtra    = 0;

	if( !RegisterClassEx(&winClass) )
		return E_FAIL;

	g_hWnd = CreateWindowEx( NULL, "WINDOWS_CLASS", 
                             "DirectX 9 - Debugging with printf in HLSL",
						     WS_OVERLAPPEDWINDOW | WS_VISIBLE,
					         0, 0, 640, 480, NULL, NULL, hInstance, NULL );

	if( g_hWnd == NULL )
		return E_FAIL;

    ShowWindow( g_hWnd, nCmdShow );
    UpdateWindow( g_hWnd );

	init();
	initEffect();

	while( uMsg.message != WM_QUIT )
	{
		if( PeekMessage( &uMsg, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &uMsg );
			DispatchMessage( &uMsg );
		}
		else
		    render();
	}

	shutDown();

    UnregisterClass( "WINDOWS_CLASS", winClass.hInstance );

	return uMsg.wParam;
}

//-----------------------------------------------------------------------------
// Name: WindowProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT CALLBACK WindowProc( HWND   hWnd, 
							 UINT   msg, 
							 WPARAM wParam, 
							 LPARAM lParam )
{
    switch( msg )
	{
        case WM_KEYDOWN:
		{
			switch( wParam )
			{
				case 'W':
					g_TestCoordinate[1] -= 1.0f/256.0f;
					break;
				case 'A':
					g_TestCoordinate[0] -= 1.0f/256.0f;
					break;
				case 'S':
					g_TestCoordinate[0] += 1.0f/256.0f;
					break;
				case 'Y':
					g_TestCoordinate[1] += 1.0f/256.0f;
					break;
				case VK_UP:
					g_XRot += (float)M_PI/100.0f;
					break;
				case VK_DOWN:
					g_XRot -= (float)M_PI/100.0f;
					break;
				case VK_LEFT:
					g_YRot += (float)M_PI/100.0f;
					break;
				case VK_RIGHT:
					g_YRot -= (float)M_PI/100.0f;
					break;
				case VK_ESCAPE:
					PostQuitMessage(0);
					break;
			}
		}
        break;

		case WM_CLOSE:
		{
			PostQuitMessage(0);	
		}
		
        case WM_DESTROY:
		{
            PostQuitMessage(0);
		}
        break;

		default:
		{
			return DefWindowProc( hWnd, msg, wParam, lParam );
		}
		break;
	}

	return 0;
}

//-----------------------------------------------------------------------------
// Name: init()
// Desc: 
//-----------------------------------------------------------------------------
void init( void )
{
    g_pD3D = Direct3DCreate9( D3D_SDK_VERSION );

    D3DDISPLAYMODE d3ddm;

    g_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm );

    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );

    d3dpp.Windowed               = TRUE;
    d3dpp.SwapEffect             = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat       = d3ddm.Format;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
    d3dpp.PresentationInterval   = D3DPRESENT_INTERVAL_IMMEDIATE;

    g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, g_hWnd,
                          D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                          &d3dpp, &g_pd3dDevice );

	g_pd3dDevice->CreateVertexBuffer( 4*sizeof(Vertex), D3DUSAGE_WRITEONLY, 
                                      D3DFVF_VERTEX, D3DPOOL_DEFAULT, 
                                      &g_pVertexBuffer, NULL );
    void *pVertices = NULL;

    g_pVertexBuffer->Lock( 0, sizeof(g_quadVertices), (void**)&pVertices, 0 );
    memcpy( pVertices, g_quadVertices, sizeof(g_quadVertices) );
    g_pVertexBuffer->Unlock();

	g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

	D3DXMatrixPerspectiveFovLH( &g_matProj, D3DXToRadian( 45.0f ), 
                                640.0f / 480.0f, 0.1f, 100.0f );
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &g_matProj );

	D3DXMatrixIdentity( &g_matView );
}


//-----------------------------------------------------------------------------
// Name: initEffect()
// Desc: Initializie an Fx effect.
//-----------------------------------------------------------------------------
void initEffect( void )
{
	//
	// Create a test texture for our effect to use...
	//

	D3DXCreateTextureFromFile( g_pd3dDevice, "decal.bmp", &g_pDecalTexture );
	D3DXCreateTextureFromFile( g_pd3dDevice, "normal.bmp", &g_pBumpTexture );
	D3DXCreateTextureFromFile( g_pd3dDevice, "font.bmp", &g_pFontTexture );

	HRESULT hr;
	LPD3DXBUFFER pBufferErrors = NULL;

	hr = D3DXCreateEffectFromFile( g_pd3dDevice, 
		                           "demodebug.fx",
		                           NULL, 
		                           NULL, 
		                           0, 
		                           NULL, 
		                           &g_pEffect, 
		                           &pBufferErrors );

	if( FAILED(hr) )
	{
		LPVOID pCompilErrors = pBufferErrors->GetBufferPointer();
		MessageBox(NULL, (const char*)pCompilErrors, "Fx Compile Error",
			MB_OK|MB_ICONEXCLAMATION);
	}
	
}

//-----------------------------------------------------------------------------
// Name: shutDown()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
void shutDown( void )
{
	if( g_pDecalTexture != NULL ) 
    {
		g_pDecalTexture->Release(); 
        g_pDecalTexture = NULL;
	}
	
	if( g_pBumpTexture != NULL ) 
    {
		g_pBumpTexture->Release(); 
        g_pBumpTexture = NULL;
	}
	
	if( g_pFontTexture != NULL ) 
    {
		g_pFontTexture->Release(); 
        g_pFontTexture = NULL;
	}
	
	if( g_pEffect != NULL )
	{
		g_pEffect->Release(); 
        g_pEffect = NULL;
	}

    if( g_pVertexBuffer != NULL ) 
    {
        g_pVertexBuffer->Release(); 
        g_pVertexBuffer = NULL; 
    }

    if( g_pd3dDevice != NULL ) 
    {
        g_pd3dDevice->Release(); 
        g_pd3dDevice = NULL; 
    }

    if( g_pD3D != NULL ) 
    {
        g_pD3D->Release(); 
        g_pD3D = NULL; 
    }
}

//-----------------------------------------------------------------------------
// Name: setTechniqueVariables()
// Desc: 
//-----------------------------------------------------------------------------
void setTechniqueVariables( void )
{
	D3DXMATRIX matRotX;
	D3DXMATRIX matRotY;
	D3DXMatrixRotationX(&matRotX, g_XRot);
	D3DXMatrixRotationY(&matRotY, g_YRot);
    D3DXMatrixTranslation( &g_matWorld, 0.0f, 0.0f, 4.0f );
	g_matWorld = matRotX * matRotY * g_matWorld;
    D3DXMatrixIdentity( &g_matView ); 

    D3DXMATRIX worldView = g_matWorld * g_matView;
	D3DXMatrixTranspose( &worldView, &worldView);

    D3DXMATRIX worldViewProjection = g_matWorld * g_matView * g_matProj;
	D3DXMatrixTranspose( &worldViewProjection, &worldViewProjection );

	g_pEffect->SetMatrix( "worldViewProj", &worldViewProjection );

	g_pEffect->SetFloatArray("lightDir", g_LightDir, 3);
	g_pEffect->SetFloatArray("testCoordinate", g_TestCoordinate, 2);
	g_pEffect->SetMatrix( "worldView", &worldView);
	g_pEffect->SetMatrix( "worldViewProj", &worldViewProjection );
	g_pEffect->SetTexture( "decalTexture", g_pDecalTexture );
	g_pEffect->SetTexture( "bumpTexture", g_pBumpTexture );
	g_pEffect->SetTexture( "fontTexture", g_pFontTexture );
}

//-----------------------------------------------------------------------------
// Name: render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
void render( void )
{
    g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                         D3DCOLOR_COLORVALUE(0.0f,0.0f,0.0f,1.0f), 1.0f, 0 );

	g_pd3dDevice->BeginScene();

	g_pEffect->SetTechnique( "DemoDebugTechnique" );

	setTechniqueVariables();

	UINT uPasses;
    g_pEffect->Begin( &uPasses, 0 );
    
    for( UINT uPass = 0; uPass < uPasses; ++uPass )
    {
        g_pEffect->BeginPass( uPass );

        g_pd3dDevice->SetStreamSource( 0, g_pVertexBuffer, 0, sizeof(Vertex) );
		g_pd3dDevice->SetFVF( D3DFVF_VERTEX );
		g_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );

        g_pEffect->EndPass();
    }
 
    g_pEffect->End();

	g_pd3dDevice->EndScene();
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}
